//**************************************************************************
//*                     This file is part of the                           *
//*                MMC - Mpxplay Multimedia Commander                      *
//*                   The source code of MMC is                            *
//*        (C) copyright 1998-2018 by PDSoft (Attila Padar)                *
//*                http://mpxplay.sourceforge.net                          *
//*                  email: mpxplay@freemail.hu                            *
//**************************************************************************
//*  This program is distributed in the hope that it will be useful,       *
//*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *
//*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                  *
//*  Please contact with the author (with me) if you want to use           *
//*  or modify this source.                                                *
//**************************************************************************
//function: video widget handling

//#define MPXPLAY_USE_DEBUGF 1
#define DISPQT_DEBUG_OUTPUT NULL //stdout
#define DISPQT_DEBUG_ERROR stderr
#define DISPQT_DEBUGOUT_FFMPEG stdout // NULL
#define DISPQT_DEBUGOUT_RENDER NULL // stdout
#define DISPQT_DEBUGOUT_SUBTITLE NULL // stdout
#define DISPQT_DEBUGOUT_CROP NULL // stdout
#define DISPQT_DEBUGOUT_ACROP NULL // stdout

#include <QtGui>
#include <QtWidgets>
#include <QtGui/QPixMap>
#include <QtPlatformHeaders/QWindowsWindowFunctions.h>
#include "moc_video_qt.h"
#include "moc_mainwindow.h"

//#if (DISPQT_USE_OPENGL_WIDGET == 1)
//extern QOpenGLContext *qt_gl_global_share_context();
//#endif

#ifdef DISPQT_USE_GDI_WIDGET
#include <windows.h>
#include <objidl.h>
#include <gdiplus.h>
#include <wingdi.h>
#endif

#define DISPQT_VIDEO_AR_SCALE 1000
#define DISPQT_VIDEO_ROUND_VALUE 1
#define DISPQT_VIDEO_ROUND_MASK (~1)

#ifdef MPXPLAY_LINK_ORIGINAL_FFMPEG

#if (DISPQT_USE_OPENGL_WIDGET == 2)
FFMpegVideoWidget::FFMpegVideoWidget(MainWindow *main_window, QWidget *parent) : QOpenGLWidget(parent)
#else
FFMpegVideoWidget::FFMpegVideoWidget(MainWindow *main_window, QWidget *parent) : QWidget(parent)
#endif
{
	struct mmc_dispqt_config_s *gcfg = main_window->gui_config;
	mpxplay_debugf(DISPQT_DEBUGOUT_FFMPEG,"FFMpegVideoWidget INIT");
	this->main_window = main_window;
	this->parent_widget = parent;
	this->sub_title_infos = NULL;

	memset(&this->video_source_infos, 0, sizeof(this->video_source_infos));

	memset(&this->video_surface_infos, 0, sizeof(this->video_surface_infos));
	this->video_surface_infos.sws_dst_width = this->video_surface_infos.window_size_x = this->width();
	this->video_surface_infos.sws_dst_height = this->video_surface_infos.window_size_y = this->height();
	this->video_surface_infos.video_zoom_globalpos_x = this->video_surface_infos.window_size_x / 2;
	this->video_surface_infos.video_zoom_globalpos_y = this->video_surface_infos.window_size_y / 2;
	this->video_surface_infos.window_pos_x = QWidget::mapToGlobal(this->pos()).x();
	this->video_surface_infos.window_pos_y = QWidget::mapToGlobal(this->pos()).y();
	this->video_surface_infos.videoout_surface_pix_fmt = DISPQT_VIDEO_QIMAGE_OUTPUT_AVFORMAT; // TODO: use separated video output config
	//this->video_surface_infos.last_video_render_type = gcfg->video_renderer_type;
	//this->video_surface_infos.video_renderer_evr = new EVRCustomPresenter(NULL);
	//this->video_surface_infos.video_sample_scheduler_evr = new Scheduler(this->video_renderer_evr);

#if (DISPQT_USE_OPENGL_WIDGET == 1)
	this->video_surface_infos.video_opengl_context = new QOpenGLContext(this);
	this->video_surface_infos.video_opengl_sfmt = new QSurfaceFormat();
	this->video_surface_infos.video_opengl_sfmt->setRenderableType(QSurfaceFormat::OpenGL);
	this->video_surface_infos.video_opengl_sfmt->setVersion(3,3);
	this->video_surface_infos.video_opengl_sfmt->setProfile(QSurfaceFormat::CoreProfile);
	this->video_surface_infos.video_opengl_sfmt->setSwapBehavior(QSurfaceFormat::SwapBehavior::DoubleBuffer);
	this->video_surface_infos.video_opengl_context->setFormat(*this->video_surface_infos.video_opengl_sfmt);
	QWidget *toplevel_window = this->window();
	this->video_surface_infos.topwindow_hnd = (toplevel_window)? toplevel_window->windowHandle() : NULL;
	if(this->video_surface_infos.topwindow_hnd)
	{
		this->video_surface_infos.topwindow_hnd->setSurfaceType(QWindow::OpenGLSurface);
		this->video_surface_infos.topwindow_hnd->setFormat(*this->video_surface_infos.video_opengl_sfmt);
		//this->video_surface_infos.video_opengl_context->setScreen(topwindow_hnd->screen());
	}
	bool created_ctx = this->video_surface_infos.video_opengl_context->create();
	bool makedcurrent = this->video_surface_infos.video_opengl_context->makeCurrent(this->video_surface_infos.topwindow_hnd);
	this->video_surface_infos.video_opengl_device = new QOpenGLPaintDevice;//(topwindow_hnd);

//	this->video_surface_infos.video_opengl_surface = new QOffscreenSurface;
//    this->video_surface_infos.video_opengl_surface->setFormat(this->video_surface_infos.video_opengl_context->format());
//    this->video_surface_infos.video_opengl_surface->setScreen(this->video_surface_infos.video_opengl_context->screen());
//    this->video_surface_infos.video_opengl_surface->create();

//    bool makedcurrent = this->video_surface_infos.video_opengl_context->makeCurrent(this->video_surface_infos.video_opengl_surface);
	//this->main_window->windowHandle()->setFormat(*this->video_surface_infos.video_opengl_sfmt);
//	if(sharedContext){
//		this->video_surface_infos.video_opengl_context->setScreen(sharedContext->screen());
//		this->video_surface_infos.video_opengl_context->setShareContext(sharedContext);
//	}
	this->video_surface_infos.video_opengl_initialized = this->video_surface_infos.video_opengl_context->isValid();
	//QWindow::SurfaceType ftype = this->main_window->windowHandle()->surfaceType();

#elif (DISPQT_USE_OPENGL_WIDGET == 2)
	QWidget *toplevel_window = this->window();
	QWindow *topwindow_hnd = (toplevel_window)? toplevel_window->windowHandle() : NULL;
	if(topwindow_hnd)
	{
		QWindowsWindowFunctions::setHasBorderInFullScreen(topwindow_hnd->windowHandle(), true);
	}
#endif

#ifdef DISPQT_USE_GDI_WIDGET
    setAttribute(Qt::WA_NativeWindow);
    setAttribute(Qt::WA_PaintOnScreen);
#endif

    this->video_sw_transform = new FFMpegVideoSoftwareTransform(main_window, &this->video_source_infos, &this->video_surface_infos);

	connect(this, SIGNAL(signal_ffmpeg_swctx_resize(int,int,bool)), this, SLOT(videowidget_ffmpeg_swctx_resize(int,int,bool)));
	connect(this, SIGNAL(signal_ffmpeg_swctx_move(int,int)), this, SLOT(videowidget_ffmpeg_swctx_move(int,int)));
	connect(this, SIGNAL(signal_ffmpeg_swctx_zoom(int,int,int)), this, SLOT(videowidget_ffmpeg_swctx_zoom(int,int,int)));
	connect(this, SIGNAL(signal_ffmpeg_filtered_trans_refresh()), this, SLOT(videowidget_ffmpeg_filtered_trans_refresh()));
	connect(this, SIGNAL(signal_ffmpeg_filtered_eq_refresh()), this, SLOT(videowidget_ffmpeg_filtered_eq_refresh()));

	this->video_decoder_worker = new FFMpegVideoDecoderWorker(main_window);
	this->video_decoder_worker->decoderworker_video_surface_refresh(DISPQT_VIDEOSURFACE_REFRESHTYPE_CLEAR); // !!! initial window CLEAR
}

FFMpegVideoWidget::~FFMpegVideoWidget()
{
	struct dispqt_video_surface_info_s *video_surface_infos = &this->video_surface_infos;
	this->mutex_video.tryLock(DISPQT_VIDEO_MUTEX_TIMEOUT);
#if (DISPQT_USE_OPENGL_WIDGET == 1)
	video_surface_infos->video_opengl_context->doneCurrent();
#endif
	if(this->video_decoder_worker){
		FFMpegVideoDecoderWorker *decoder_worker = this->video_decoder_worker;
		this->video_decoder_worker = NULL;
		delete decoder_worker;
	}
	if(this->video_sw_transform)
	{
		delete this->video_sw_transform;
	}
	videowidget_ffmpeg_subtitle_close();
	this->mutex_video.unlock();
}

//bool FFMpegVideoWidget::event(QEvent *event)
//{
//	switch(event->type())
//	{
//		case QEvent::Expose:
//			event->accept();
//			return true;
//	}
//	return QWidget::event(event);
//}

// --------------------------------------------------------------------------------------------------------------------

void FFMpegVideoWidget::videowidget_ffmpeg_swctx_resize(int window_width, int window_height, bool do_repaint)
{
	struct dispqt_video_surface_info_s *video_surface_infos = &this->video_surface_infos;
	video_surface_infos->updateresize_x = window_width;
	video_surface_infos->updateresize_y = window_height;
	video_surface_infos->do_resize_repaint = do_repaint;
	video_surface_infos->update_resize = true;
	this->video_decoder_worker->decoderworker_video_surface_refresh(DISPQT_VIDEOSURFACE_REFRESHTYPE_UPDATE);
}

void FFMpegVideoWidget::videowidget_ffmpeg_swctx_move(int move_x, int move_y)
{
//	struct dispqt_video_surface_info_s *video_surface_infos = &this->video_surface_infos;
//	video_surface_infos->updatemove_x += move_x;
//	video_surface_infos->updatemove_y += move_y;
//	video_surface_infos->update_move = true;
//	this->video_decoder_worker->decoderworker_video_surface_refresh(DISPQT_VIDEOSURFACE_REFRESHTYPE_UPDATE);
	if(videowidget_ffmpeg_surface_move_calculate(move_x, move_y))
		this->videowidget_ffmpeg_filtered_surface_update(false);
}

bool FFMpegVideoWidget::videowidget_ffmpeg_surface_move_calculate(int move_x, int move_y)
{
	struct dispqt_video_source_info_s *video_source_infos = &this->video_source_infos;
	struct dispqt_video_surface_info_s *video_surface_infos = &this->video_surface_infos;
	struct mmc_dispqt_config_s *gcfg = this->main_window->gui_config;
	int max_crop_x, max_crop_y, prev_crop_x, prev_crop_y;
	bool do_repaint = false;

	prev_crop_x = video_surface_infos->sws_crop_x;
	prev_crop_y = video_surface_infos->sws_crop_y;

	move_x = move_x * DISPQT_CONFIG_VIDEOZOOM_SCALE / gcfg->video_zoom_value;  // scale moving to zoom rate
	max_crop_x = video_source_infos->sws_srccrop_width - video_surface_infos->sws_crop_w;
	if(max_crop_x > 0)
	{
		video_surface_infos->sws_crop_x += move_x;
		if(video_surface_infos->sws_crop_x < video_source_infos->sws_srccrop_x)
			video_surface_infos->sws_crop_x = video_source_infos->sws_srccrop_x;
		else if(video_surface_infos->sws_crop_x > (video_source_infos->sws_srccrop_x + max_crop_x))
			video_surface_infos->sws_crop_x = video_source_infos->sws_srccrop_x + max_crop_x;
	}

	move_y = move_y * DISPQT_CONFIG_VIDEOZOOM_SCALE / gcfg->video_zoom_value;
	max_crop_y = video_source_infos->sws_srccrop_height - video_surface_infos->sws_crop_h;
	if(max_crop_y > 0)
	{
		video_surface_infos->sws_crop_y += move_y;
		if(video_surface_infos->sws_crop_y < video_source_infos->sws_srccrop_y)
			video_surface_infos->sws_crop_y = video_source_infos->sws_srccrop_y;
		else if(video_surface_infos->sws_crop_y > (video_source_infos->sws_srccrop_y + max_crop_y))
			video_surface_infos->sws_crop_y = video_source_infos->sws_srccrop_y + max_crop_y;
	}

	// reconfig zoom center after move
	video_surface_infos->video_zoom_globalpos_x = video_surface_infos->sws_dst_width / 2;
	video_surface_infos->video_zoom_globalpos_y = video_surface_infos->sws_dst_height / 2;
	if(!this->isFullScreen()){
		video_surface_infos->video_zoom_globalpos_x += QWidget::mapToGlobal(this->pos()).x();
		video_surface_infos->video_zoom_globalpos_y += QWidget::mapToGlobal(this->pos()).y();
	}

	if((prev_crop_x != video_surface_infos->sws_crop_x) || (prev_crop_y = video_surface_infos->sws_crop_y)){
		mpxplay_debugf(DISPQT_DEBUGOUT_FFMPEG, "MOVE cx:%d cy:%d", video_surface_infos->sws_crop_x, video_surface_infos->sws_crop_y);
		do_repaint = true;
	}
	return do_repaint;
}

void FFMpegVideoWidget::videowidget_ffmpeg_swctx_zoom(int globalpos_x, int globalpos_y, int direction)
{
//	struct dispqt_video_surface_info_s *video_surface_infos = &this->video_surface_infos;
//	video_surface_infos->updatezoom_center_x = globalpos_x;
//	video_surface_infos->updatezoom_center_y = globalpos_y;
//	video_surface_infos->updatezoom_direction += direction;
//	video_surface_infos->update_zoom = true;
//	this->video_decoder_worker->decoderworker_video_surface_refresh(DISPQT_VIDEOSURFACE_REFRESHTYPE_UPDATE);
	if(videowidget_ffmpeg_surface_zoom_calculate(globalpos_x, globalpos_y, direction))
		this->videowidget_ffmpeg_filtered_surface_update(false);
	emit this->main_window->signal_video_mouse_cursor_show(true);
}

bool FFMpegVideoWidget::videowidget_ffmpeg_surface_zoom_calculate(int globalpos_x, int globalpos_y, int direction)
{
	struct dispqt_video_source_info_s *video_source_infos = &this->video_source_infos;
	struct dispqt_video_surface_info_s *video_surface_infos = &this->video_surface_infos;
	struct mmc_dispqt_config_s *gcfg = this->main_window->gui_config;
	int i, zoom_val, center_values[2];
	bool do_update = false;

	if((video_surface_infos->last_video_zoom_min_val <= 0) || (video_surface_infos->last_video_zoom_ti_val <= 0) || (video_source_infos->sws_srccrop_width <= 0) || (video_source_infos->sws_srccrop_height <= 0) || (gcfg->video_zoom_value <= 0))
		return do_update;

	zoom_val = gcfg->video_zoom_value + direction * DISPQT_CONFIG_VIDEOZOOM_STEP * gcfg->video_zoom_value / DISPQT_CONFIG_VIDEOZOOM_SCALE;

	center_values[0] = DISPQT_CONFIG_VIDEOZOOM_CENTER;
	center_values[1] = video_surface_infos->last_video_zoom_ti_val;
	mpxplay_debugf(DISPQT_DEBUGOUT_CROP, "swctx_zoom cv:%d zv:%d mv:%d", gcfg->video_zoom_value, zoom_val, video_surface_infos->last_video_zoom_min_val);

	if(zoom_val < video_surface_infos->last_video_zoom_min_val)
		zoom_val = video_surface_infos->last_video_zoom_min_val;
	else if(zoom_val > DISPQT_CONFIG_VIDEOZOOM_MAX)
		zoom_val = DISPQT_CONFIG_VIDEOZOOM_MAX;

	for(i = 0; i < (sizeof(center_values) / sizeof(center_values[0])); i++)
	{
		const int center = center_values[i];
		//mpxplay_debugf(DISPQT_DEBUGOUT_CROP, "swctx_zoom i:%d c:%d", i, center);
		if(((gcfg->video_zoom_value > center) && (zoom_val <= (center + center / 10))) || ((gcfg->video_zoom_value < center) && (zoom_val >= (center - center / 10)))){
			zoom_val = center;
			mpxplay_debugf(DISPQT_DEBUGOUT_CROP, "swctx_zoom i:%d c:%d", i, center);
			break;
		}
	}

	mpxplay_debugf(DISPQT_DEBUGOUT_CROP, "swctx_zoom zt:%d zv:%d mv:%d", gcfg->video_zoom_type, zoom_val, video_surface_infos->last_video_zoom_min_val);

	if(gcfg->video_zoom_value != zoom_val)
	{
		gcfg->video_zoom_type = DISPQT_VIDEOFULLSCREENZOOMTYPE_MANUAL;
		gcfg->video_zoom_value = zoom_val;
		if(globalpos_x >= 0)
			video_surface_infos->video_zoom_globalpos_x = globalpos_x;
		if(globalpos_y >= 0)
			video_surface_infos->video_zoom_globalpos_y = globalpos_y;
		this->video_surface_infos.videoout_surface_recalc = true;
		do_update = true;
	}
	mpxplay_debugf(DISPQT_DEBUGOUT_RENDER, "videowidget_ffmpeg_swctx_zoom END 1");
	return do_update;
}

void FFMpegVideoWidget::resizeEvent(QResizeEvent *e)
{
	this->videowidget_ffmpeg_swctx_resize(e->size().width(), e->size().height(), false);
#if (DISPQT_USE_OPENGL_WIDGET == 2)
	QOpenGLWidget::resizeEvent(e);
#elif !defined(DISPQT_USE_GDI_WIDGET)
	QWidget::resizeEvent(e);
#endif
}

void FFMpegVideoWidget::videowidget_video_subtitle_render(QPainter *painter, ffmpegvideo_subtitle_info_s *subtitle_infos)
{
	struct mmc_dispqt_config_s *gcfg = this->main_window->gui_config;
	struct dispqt_video_surface_info_s *video_surface_infos = &this->video_surface_infos;
	int i, scale, charpixelsize, x, y, cy, cx;
	unsigned int flags;
#ifdef DISPQT_FFMPEG_TEST_SPEED
	mpxp_int64_t render_time = pds_gettimem();
#endif

	if(!painter || !subtitle_infos)
		return;

	if(subtitle_infos->last_size_x && subtitle_infos->last_size_y)
	{
		// clear black subtitle areas
		if(subtitle_infos->last_render_x < video_surface_infos->sws_dst_pos_x)
			painter->fillRect(0, 0, video_surface_infos->sws_dst_pos_x, video_surface_infos->window_size_y, Qt::black); // left side
		cx = video_surface_infos->sws_dst_pos_x + video_surface_infos->sws_dst_width;
		if((video_surface_infos->window_size_x > cx) && ((subtitle_infos->last_render_x + subtitle_infos->last_size_x) > cx))
			painter->fillRect(cx, 0, video_surface_infos->window_size_x - cx, video_surface_infos->window_size_y, Qt::black);  // right side
		if(subtitle_infos->last_render_y < video_surface_infos->sws_dst_pos_y)
			painter->fillRect(0, 0, video_surface_infos->window_size_x, video_surface_infos->sws_dst_pos_y, Qt::black);   // top side
		cy = video_surface_infos->sws_dst_pos_y + video_surface_infos->sws_dst_height;
		if((video_surface_infos->window_size_y > cy) && ((subtitle_infos->last_render_y + subtitle_infos->last_size_y) > cy))
			painter->fillRect(0, cy, video_surface_infos->window_size_x, video_surface_infos->window_size_y - cy, Qt::black);  // bottom side
	}

	subtitle_infos->last_render_x = subtitle_infos->last_render_y = subtitle_infos->last_size_x = subtitle_infos->last_size_y = 0;

	if(!subtitle_infos->nb_lines)
		return;

	scale = 100 * video_surface_infos->window_size_y / QApplication::desktop()->screenGeometry().height();
	if(scale < 5)
		scale = 5;
	else if(scale > 100)
		scale = 100;

	if(gcfg->font_text_name[DISPQT_FONTTEXTID_SUBTITLE])
		this->font_subtitle.fromString(QString::fromUtf8(gcfg->font_text_name[DISPQT_FONTTEXTID_SUBTITLE])); // TODO: move out from here
	charpixelsize = this->font_subtitle.pointSize();
	charpixelsize = charpixelsize * scale / 100;
	this->font_subtitle.setPointSize(charpixelsize);
	painter->setFont(this->font_subtitle);

	y = video_surface_infos->window_size_y - subtitle_infos->nb_lines * charpixelsize * 3 / 2 - charpixelsize * 2;
	if(y < 0)
		y = 0;
	subtitle_infos->last_render_y = y;
	flags = 0;

	for(i = 0; i < subtitle_infos->nb_lines; i++)
	{
		QFontMetrics fm(this->font_subtitle);
		int textpixellen = fm.width(QString::fromUtf8(subtitle_infos->line_texts[i]));
		if(textpixellen <= 0)
			continue;
		x = video_surface_infos->window_size_x / 2 - textpixellen / 2;
		if(x < 0)
			x = 0;
		if(flags != subtitle_infos->line_flags[i]){
			flags = subtitle_infos->line_flags[i];
			this->font_subtitle.setItalic((flags & INFFMPEG_ASS_DECODED_FLAG_ITALIC)? true : false);
			painter->setFont(this->font_subtitle);
		}
		painter->setPen(Qt::black);
		painter->drawText(x + 2, y + 2, QString::fromUtf8(subtitle_infos->line_texts[i])); // shadow

		painter->setPen(Qt::white);
		painter->drawText(x, y, QString::fromUtf8(subtitle_infos->line_texts[i])); // white text

		//mpxplay_debugf(DISPQT_DEBUGOUT_SUBTITLE, "SUBTITLE drawText: w:%d cs:%d x:%d y:%d l:%d cy:%d %s", video_surface_infos->window_size_x, charpixelsize, x, y, i, cy, subtitle_infos->line_texts[i]);
		y += charpixelsize * 3 / 2;
		subtitle_infos->last_render_x = (i == 0)? x : min(x, subtitle_infos->last_render_x);
		subtitle_infos->last_size_x = max(subtitle_infos->last_size_x, textpixellen);
	}
	subtitle_infos->last_size_y = y - subtitle_infos->last_render_y;
#ifdef DISPQT_FFMPEG_TEST_SPEED
	render_time = pds_gettimem() - render_time;
	if(render_time > 0){
		//mpxplay_debugf(DISPQT_DEBUGOUT_RENDER, "videowidget_video_subtitle_render TIME %llu", render_time);
	}
#endif
}

//-------------------------------------------------------------------------------------------------------------------------------------

#ifdef DISPQT_USE_GDI_WIDGET
QPaintEngine* FFMpegVideoWidget::paintEngine() const
{
    return NULL;
}

void FFMpegVideoWidget::paintEvent(QPaintEvent *event)
{
}

void FFMpegVideoWidget::videowidget_video_gdi_render(struct dispqt_video_surface_info_s *video_surface_infos)
{
	struct dispqt_video_filter_info_s *video_filter_infos = &this->video_filter_infos;
	HWND hwnd = (HWND)this->winId();
	HDC hDC = GetDC(hwnd);

	if(video_filter_infos->videoout_surface_clearpic || !video_filter_infos->ffmpeg_output_frame)
	{
	    RECT rect;
	    GetClientRect(hwnd, &rect);
	    HBRUSH hbrRed = CreateSolidBrush(RGB(0, 0, 0));
	    FillRect(hDC, &rect, hbrRed);
		video_filter_infos->videoout_surface_clearpic = false;
	}

	if(video_filter_infos->ffmpeg_output_frame)
	{
		HDC hMemDC = CreateCompatibleDC(hDC);
		HBITMAP hOldBMP, hNewBMP;
		//BITMAP bm;

		hNewBMP = CreateBitmap(video_filter_infos->ffmpeg_output_frame->width, video_filter_infos->ffmpeg_output_frame->height, 1, 32, video_filter_infos->ffmpeg_output_frame->data[0]);

		hOldBMP = (HBITMAP)SelectObject(hMemDC, hNewBMP);

		//int result = GetObject(hNewBMP, sizeof(bm), &bm);

		SetStretchBltMode(hDC, COLORONCOLOR); // HALFTONE

		StretchBlt(hDC, video_surface_infos->sws_dst_pos_x, video_surface_infos->sws_dst_pos_y, video_surface_infos->sws_dst_width, video_surface_infos->sws_dst_height,
		           hMemDC, video_surface_infos->sws_crop_x, video_surface_infos->sws_crop_y, video_surface_infos->sws_crop_w, video_surface_infos->sws_crop_h, SRCCOPY);
		//BitBlt(hDC, 0, 0, video_filter_infos->ffmpeg_output_frame->width, video_filter_infos->ffmpeg_output_frame->height, hMemDC, 0, 0, SRCCOPY);

		SelectObject(hMemDC, hOldBMP);

		DeleteObject(hNewBMP);
		DeleteDC(hMemDC);
	}

//    QString text("Test GDI Paint");
//    RECT rect;
//    GetClientRect(hwnd, &rect);
//
//    HBRUSH hbrRed = CreateSolidBrush(RGB(0, 255, 0));
//    FillRect(hDC, &rect, hbrRed);
//
//    Ellipse(hDC, 50, 50, rect.right - 100, rect.bottom - 100);
//    SetTextAlign(hDC, TA_CENTER | TA_BASELINE);
//    TextOutW(hDC, width() / 2, height() / 2, (LPCWSTR)text.utf16(), text.size());
    ReleaseDC(hwnd, hDC);
}

#else

void FFMpegVideoWidget::videowidget_video_frame_render(QPainter *p)
{
	struct dispqt_video_surface_info_s *video_surface_infos = &this->video_surface_infos;

	if(video_surface_infos->videoout_surface_clearpic || !video_surface_infos->ffmpeg_output_frame)
	{
		p->fillRect(0, 0, video_surface_infos->window_size_x, video_surface_infos->window_size_y, Qt::black);
		video_surface_infos->videoout_surface_clearpic = false;
		//mpxplay_debugf(DISPQT_DEBUGOUT_RENDER, "paintEvent CLEARPIC");
	}
	if(video_surface_infos->ffmpeg_output_frame)
	{
		p->setRenderHint(QPainter::SmoothPixmapTransform, true);
		QImage image(video_surface_infos->ffmpeg_output_frame->data[0], video_surface_infos->ffmpeg_output_frame->width, video_surface_infos->ffmpeg_output_frame->height, DISPQT_VIDEO_QIMAGE_OUTPUT_QFORMAT);
		p->drawImage(QRectF(video_surface_infos->sws_dst_pos_x, video_surface_infos->sws_dst_pos_y, video_surface_infos->sws_dst_width, video_surface_infos->sws_dst_height), image,
				QRectF(video_surface_infos->sws_crop_x, video_surface_infos->sws_crop_y, video_surface_infos->sws_crop_w, video_surface_infos->sws_crop_h),
				Qt::NoFormatConversion);
//		QPixmap pix_map();
//		p->drawPixmap(QRectF(video_surface_infos->sws_dst_pos_x, video_surface_infos->sws_dst_pos_y, video_surface_infos->sws_dst_width, video_surface_infos->sws_dst_height), pix_map,
//				QRectF(video_surface_infos->sws_crop_x, video_surface_infos->sws_crop_y, video_surface_infos->sws_crop_w, video_surface_infos->sws_crop_h));
	}
}

void FFMpegVideoWidget::paintEvent(QPaintEvent *event)
{
	struct dispqt_video_surface_info_s *video_surface_infos = &this->video_surface_infos;
#if (DISPQT_USE_OPENGL_WIDGET == 1)
//	if(!video_surface_infos->video_opengl_initialized)
//	{
//		video_surface_infos->video_opengl_context->makeCurrent(video_surface_infos->video_opengl_context->surface());
//		video_surface_infos->video_opengl_initialized = true;
//	}
	//QOpenGLPaintDevice device(size() * devicePixelRatio());
	this->video_surface_infos.video_opengl_device->setSize(size());
	//this->video_surface_infos.video_opengl_device->setDevicePixelRatio(devicePixelRatio());
	QPainter p(this->video_surface_infos.video_opengl_device);
//#elif (DISPQT_USE_OPENGL_WIDGET == 2)
//    QOpenGLPaintDevice device(size() * devicePixelRatio());
//    QPainter p(&device);
#else
	QPainter p;
#endif
	p.begin(this);
	this->videowidget_video_frame_render(&p);
	this->videowidget_video_subtitle_render(&p, this->sub_title_infos);
	p.end();
#if (DISPQT_USE_OPENGL_WIDGET == 1)
	//video_surface_infos->video_opengl_context->doneCurrent();
	video_surface_infos->video_opengl_context->swapBuffers(this->video_surface_infos.topwindow_hnd);
//#elif (DISPQT_USE_OPENGL_WIDGET == 2)
//	QOpenGLWidget::paintEvent(event);
#endif
}

#endif

//-------------------------------------------------------------------------------------------------------------------------------------
// common video surface functions

void FFMpegVideoWidget::videowidget_ffmpeg_subtitle_store(ffmpegvideo_subtitle_info_s *subtitle_infos)
{
	if(this->sub_title_infos)
	{
		mpxplay_video_ffmpeg_subtitleinfo_clear(this->sub_title_infos);
	}
	if(subtitle_infos)
	{
		if(!this->sub_title_infos)
			this->sub_title_infos = (ffmpegvideo_subtitle_info_s *)av_mallocz(sizeof(ffmpegvideo_subtitle_info_s));
		mpxplay_video_ffmpeg_subtitleinfo_copy(this->sub_title_infos, subtitle_infos);
	}
}

void FFMpegVideoWidget::videowidget_ffmpeg_subtitle_close(void)
{
	if(this->sub_title_infos)
	{
		mpxplay_video_ffmpeg_subtitleinfo_clear(this->sub_title_infos);
		av_freep(&this->sub_title_infos);
	}
}

//-------------------------------------------------------------------------------------------------------------------------------------
void FFMpegVideoWidget::videowidget_ffmpeg_filtered_crop_refresh(bool check_autocrop)
{
	struct dispqt_video_source_info_s *video_source_infos = &this->video_source_infos;
	video_source_infos->update_crop = true;
	video_source_infos->check_autocrop = check_autocrop;
	this->video_decoder_worker->decoderworker_video_surface_refresh(DISPQT_VIDEOSURFACE_REFRESHTYPE_UPDATE);
}

void FFMpegVideoWidget::videowidget_ffmpeg_filtered_trans_refresh(void)
{
	this->video_surface_infos.update_filter_trans = true;
	this->video_decoder_worker->decoderworker_video_surface_refresh(DISPQT_VIDEOSURFACE_REFRESHTYPE_UPDATE);
}

void FFMpegVideoWidget::videowidget_ffmpeg_filtered_eq_refresh(void)
{
	this->video_surface_infos.update_filter_eq = true;
	this->video_decoder_worker->decoderworker_video_surface_refresh(DISPQT_VIDEOSURFACE_REFRESHTYPE_UPDATE);
}

void FFMpegVideoWidget::videowidget_ffmpeg_filtered_surface_update(bool immediate_paint)
{
	struct dispqt_video_surface_info_s *video_surface_infos = &this->video_surface_infos;
	AVFrame *output_frame = video_surface_infos->ffmpeg_output_frame;
	if(output_frame && ((video_surface_infos->video_outframe_last_width != output_frame->width) || (video_surface_infos->video_outframe_last_height != output_frame->height)))
	{
		video_surface_infos->video_outframe_last_width = output_frame->width;
		video_surface_infos->video_outframe_last_height = output_frame->height;
		video_surface_infos->videoout_surface_recalc = true;
	}
	if(video_surface_infos->videoout_surface_recalc)
	{
		this->videowidget_ffmpeg_swsctx_calculate(-1, -1);
	}
#ifdef DISPQT_USE_GDI_WIDGET
	this->videowidget_video_gdi_render(&this->video_surface_infos);
#else
	if(immediate_paint) // FIXME: logic is for DISPQT_VIDEORENDERTYPE_WIDGET only! (EVR has to be always immediate)
		this->repaint();
	else
		this->update();
#endif
}

//-------------------------------------------------------------------------------------------------------------------------------------
static int videowidget_ffmpeg_swsctx_get_custom_ar(char *custom_str, int scale_value, int default_value)
{
	int ar_value = default_value;
	char *ar_separator = pds_strchr(custom_str, DISPQT_CONFIG_VIDEOAR_DIVISORCHAR);
	if(ar_separator){
		int ar_custom_x = pds_atol(custom_str);
		int ar_custom_y = pds_atol(ar_separator + 1);
		if((ar_custom_x > 0) && (ar_custom_y > 0)){
			int ar = scale_value * ar_custom_x / ar_custom_y;
			if((ar >= (scale_value / 100)) && (ar <= (scale_value * 100))) // FIXME: / 100
				ar_value = ar;
		}
	}
	return ar_value;
}

//-------------------------------------------------------------------------------------------------------------------------------------

bool FFMpegVideoWidget::videowidget_ffmpeg_swsctx_calculate(int window_width, int window_height)
{
	struct dispqt_video_source_info_s *video_source_infos = &this->video_source_infos;
	struct dispqt_video_surface_info_s *video_surface_infos = &this->video_surface_infos;
	struct mmc_dispqt_config_s *gcfg = this->main_window->gui_config;
	AVFrame *output_frame;
	int video_ar, video_crop_ar = 0, window_ar, widget_global_pos_x, widget_global_pos_y;
	unsigned long ar_type;

#ifdef DISPQT_USE_GDI_WIDGET
    if(this->isFullScreen())
    {
    	window_width = QApplication::desktop()->screenGeometry().width();
    	window_height = QApplication::desktop()->screenGeometry().height();
    }
    else
#endif
    {
		if(window_width < 0)
			window_width = this->width();
		if(window_height < 0)
			window_height = this->height();
	}
	video_surface_infos->window_size_x = window_width;
	video_surface_infos->window_size_y = window_height;

	video_surface_infos->sws_dst_width = video_surface_infos->window_size_x & DISPQT_VIDEO_ROUND_MASK;
	video_surface_infos->sws_dst_height = video_surface_infos->window_size_y & DISPQT_VIDEO_ROUND_MASK;

	if((video_surface_infos->sws_dst_width <= 0) || (video_surface_infos->sws_dst_height <= 0))
		return false;

	output_frame = video_surface_infos->ffmpeg_output_frame;
	if(!output_frame || (output_frame->width <= 0) || (output_frame->height <= 0))
		return false;

	video_ar = DISPQT_VIDEO_AR_SCALE;
	video_ar *= output_frame->width * ((video_source_infos->video_aspect_ratio.num > 0)? video_source_infos->video_aspect_ratio.num : 1);
	video_ar /= output_frame->height * ((video_source_infos->video_aspect_ratio.den > 0)? video_source_infos->video_aspect_ratio.den : 1);

	// --- crop ratio calculate (manual, auto) -------------------------------------------------------------------------------------

	if((gcfg->video_crop_type != DISPQT_VIDEOSCRCROPTYPE_AUTO) || !video_source_infos->sws_srccrop_width || !video_source_infos->sws_srccrop_height
	 || (video_source_infos->sws_srccrop_width > output_frame->width) || (video_source_infos->sws_srccrop_height > output_frame->height)
	){
		video_source_infos->sws_srccrop_x = video_source_infos->sws_srccrop_y = 0;
		video_source_infos->sws_srccrop_width = output_frame->width;
		video_source_infos->sws_srccrop_height = output_frame->height;
	}

	switch(gcfg->video_crop_type)
	{
		case DISPQT_VIDEOSCRCROPTYPE_AUTO:  video_ar = video_ar * video_source_infos->sws_srccrop_width / output_frame->width * output_frame->height / video_source_infos->sws_srccrop_height; break;
		case DISPQT_VIDEOSCRCROPTYPE_4_3:   video_crop_ar = DISPQT_VIDEO_AR_SCALE * 4 / 3; break;
		case DISPQT_VIDEOSCRCROPTYPE_16_9:  video_crop_ar = DISPQT_VIDEO_AR_SCALE * 16 / 9; break;
		case DISPQT_VIDEOSCRCROPTYPE_235_1: video_crop_ar = DISPQT_VIDEO_AR_SCALE * 235 / 100; break;
		case DISPQT_VIDEOSCRCROPTYPE_CUSTOM:video_crop_ar = videowidget_ffmpeg_swsctx_get_custom_ar(gcfg->video_crop_custom, DISPQT_VIDEO_AR_SCALE, 0); break;
	}

	if((video_crop_ar > 0) && (video_crop_ar != video_ar))
	{
		if(video_ar > video_crop_ar)
			video_source_infos->sws_srccrop_width = output_frame->width * video_crop_ar / video_ar;
		else
			video_source_infos->sws_srccrop_height = output_frame->height * video_ar / video_crop_ar;
		video_source_infos->sws_srccrop_x = (output_frame->width - video_source_infos->sws_srccrop_width) / 2;
		video_source_infos->sws_srccrop_y = (output_frame->height - video_source_infos->sws_srccrop_height) / 2;
		mpxplay_debugf(DISPQT_DEBUGOUT_CROP, "CROP var:%d car:%d sw:%d sh:%d cw:%d ch:%d", video_ar, video_crop_ar, output_frame->width, output_frame->height, video_source_infos->sws_srccrop_width, video_source_infos->sws_srccrop_height);
		video_ar = video_crop_ar;
	}

	// --- aspect ratio calculate -------------------------------------------------------------------------------------

	if(video_source_infos->ffmpeg_decoded_frame && (video_source_infos->ffmpeg_decoded_frame->flags & DISPQT_AVFRAME_FLAG_AR_WINDOW)) // window (fill) aspect ratio by decoder (eg. visualization output)
	{
		ar_type = DISPQT_VIDEOFULLSCREENARTYPE_FILL;
	}
	else
	{
		ar_type = (this->isFullScreen() || (gcfg->video_windowed_ar_type == DISPQT_VIDEOWINDOWEDARTYPE_EQFULLSCREEN))? gcfg->video_fullscreen_ar_type : gcfg->video_windowed_ar_type;
	}

	window_ar = DISPQT_VIDEO_AR_SCALE * window_width / window_height;

	switch(ar_type){
		case DISPQT_VIDEOFULLSCREENARTYPE_FILL:  video_ar = window_ar; break;
		case DISPQT_VIDEOFULLSCREENARTYPE_1_1:   video_ar = DISPQT_VIDEO_AR_SCALE; break;
		case DISPQT_VIDEOFULLSCREENARTYPE_4_3:   video_ar = DISPQT_VIDEO_AR_SCALE * 4 / 3; break;
		case DISPQT_VIDEOFULLSCREENARTYPE_16_9:  video_ar = DISPQT_VIDEO_AR_SCALE * 16 / 9; break;
		case DISPQT_VIDEOFULLSCREENARTYPE_235_1: video_ar = DISPQT_VIDEO_AR_SCALE * 235 / 100; break;
		case DISPQT_VIDEOFULLSCREENARTYPE_CUSTOM:video_ar = videowidget_ffmpeg_swsctx_get_custom_ar(gcfg->video_ar_custom, DISPQT_VIDEO_AR_SCALE, DISPQT_VIDEO_AR_SCALE); break;
		case DISPQT_VIDEOFULLSCREENARTYPE_PIXEL: video_ar = DISPQT_VIDEO_AR_SCALE * output_frame->width / output_frame->height; break;
	}

	// --- zoom value calculate -----------------------------------------------------------------------------------------

	const int prev_crop_w = video_surface_infos->sws_crop_w;
	const int prev_crop_h = video_surface_infos->sws_crop_h;
	const int pixel_ar = DISPQT_VIDEO_AR_SCALE * video_source_infos->sws_srccrop_width / video_source_infos->sws_srccrop_height;

	window_width = video_source_infos->sws_srccrop_width;
	window_height = video_source_infos->sws_srccrop_height;

	if(video_ar > pixel_ar)
		window_height = window_height * pixel_ar / video_ar;
	else if((video_ar > 0) && (video_ar < pixel_ar))
		window_width = window_width * video_ar / pixel_ar;

	if((window_width <= 0) || (window_height <= 0))
		return false;

	const int zoom_val_ti_w = (DISPQT_CONFIG_VIDEOZOOM_SCALE * video_surface_infos->sws_dst_width + (window_width >> 1)) / window_width;
	const int zoom_val_ti_h = (DISPQT_CONFIG_VIDEOZOOM_SCALE * video_surface_infos->sws_dst_height + (window_height >> 1)) / window_height;
	const int zoom_val_to = max(zoom_val_ti_w , zoom_val_ti_h);
	const int zoom_val_ti = (ar_type == DISPQT_VIDEOFULLSCREENARTYPE_FILL)? zoom_val_to : min(zoom_val_ti_w , zoom_val_ti_h);
	int zoom_value = DISPQT_CONFIG_VIDEOZOOM_CENTER;

	video_surface_infos->last_video_zoom_ti_val = zoom_val_ti;
//	this->last_video_zoom_min_val = ((ar_type == DISPQT_VIDEOFULLSCREENARTYPE_FILL) || (zoom_val_ti < DISPQT_CONFIG_VIDEOZOOM_MIN))? zoom_val_ti : DISPQT_CONFIG_VIDEOZOOM_MIN; // FIXME: if video resolution is larger than window, zoom is incorrect in fill mode below zoom_val_ti
	video_surface_infos->last_video_zoom_min_val = (zoom_val_ti < DISPQT_CONFIG_VIDEOZOOM_MIN)? zoom_val_ti : DISPQT_CONFIG_VIDEOZOOM_MIN;

	switch(gcfg->video_zoom_type)
	{
		case DISPQT_VIDEOFULLSCREENZOOMTYPE_TOUCHINSIDE: zoom_value = zoom_val_ti; break;
		case DISPQT_VIDEOFULLSCREENZOOMTYPE_TOUCHOUTSIDE: zoom_value = zoom_val_to; break;
		case DISPQT_VIDEOFULLSCREENZOOMTYPE_2_1: zoom_value = 2 * DISPQT_CONFIG_VIDEOZOOM_SCALE; break;
		case DISPQT_VIDEOFULLSCREENZOOMTYPE_4_1: zoom_value = 4 * DISPQT_CONFIG_VIDEOZOOM_SCALE; break;
		case DISPQT_VIDEOFULLSCREENZOOMTYPE_CUSTOM: zoom_value = videowidget_ffmpeg_swsctx_get_custom_ar(gcfg->video_zoom_custom, DISPQT_CONFIG_VIDEOZOOM_SCALE, DISPQT_CONFIG_VIDEOZOOM_SCALE); break;
		case DISPQT_VIDEOFULLSCREENZOOMTYPE_MANUAL: zoom_value = gcfg->video_zoom_value; break;
	}

	if(zoom_value < video_surface_infos->last_video_zoom_min_val)
	{
		zoom_value = video_surface_infos->last_video_zoom_min_val;
	}
	else if(zoom_value > DISPQT_CONFIG_VIDEOZOOM_MAX)
	{
		zoom_value = DISPQT_CONFIG_VIDEOZOOM_MAX;
	}

	if(zoom_value != DISPQT_CONFIG_VIDEOZOOM_CENTER){
		window_width = window_width * zoom_value / DISPQT_CONFIG_VIDEOZOOM_SCALE;
		window_height = window_height * zoom_value / DISPQT_CONFIG_VIDEOZOOM_SCALE;
	}

	//window_width = (window_width + DISPQT_VIDEO_ROUND_VALUE) & DISPQT_VIDEO_ROUND_MASK;
	//window_height = (window_height + DISPQT_VIDEO_ROUND_VALUE) & DISPQT_VIDEO_ROUND_MASK;

	video_surface_infos->sws_crop_w = video_source_infos->sws_srccrop_width;
	video_surface_infos->sws_crop_h = video_source_infos->sws_srccrop_height;

	if(window_width > video_surface_infos->sws_dst_width){
		video_surface_infos->sws_crop_w = video_source_infos->sws_srccrop_width * video_surface_infos->sws_dst_width / window_width;
		window_width = video_surface_infos->sws_dst_width;
	}
	if(window_height > video_surface_infos->sws_dst_height){
		video_surface_infos->sws_crop_h = video_source_infos->sws_srccrop_height * video_surface_infos->sws_dst_height / window_height;
		window_height = video_surface_infos->sws_dst_height;
	}

	//this->sws_crop_w = (this->sws_crop_w + DISPQT_VIDEO_ROUND_VALUE) & DISPQT_VIDEO_ROUND_MASK;
	//this->sws_crop_h = (this->sws_crop_h + DISPQT_VIDEO_ROUND_VALUE) & DISPQT_VIDEO_ROUND_MASK;

	gcfg->video_zoom_value = zoom_value;

	// --- zoom center (source view window) calculate -----------------------------------------------------------------------------------------

	if(this->isFullScreen()){
		widget_global_pos_x = widget_global_pos_y = 0;
	}else{
		widget_global_pos_x = QWidget::mapToGlobal(this->pos()).x();
		widget_global_pos_y = QWidget::mapToGlobal(this->pos()).y();
	}

	int max_crop_x = video_source_infos->sws_srccrop_width - video_surface_infos->sws_crop_w;
	int max_crop_y = video_source_infos->sws_srccrop_height - video_surface_infos->sws_crop_h;

	if( (gcfg->video_zoom_type == DISPQT_VIDEOFULLSCREENZOOMTYPE_TOUCHINSIDE) || (gcfg->video_zoom_type == DISPQT_VIDEOFULLSCREENZOOMTYPE_TOUCHOUTSIDE)){
		video_surface_infos->video_zoom_globalpos_x = widget_global_pos_x + video_surface_infos->sws_dst_width / 2;
		video_surface_infos->video_zoom_globalpos_y = widget_global_pos_y + video_surface_infos->sws_dst_height / 2;
		video_surface_infos->sws_crop_x = video_source_infos->sws_srccrop_x + max_crop_x / 2;
		video_surface_infos->sws_crop_y = video_source_infos->sws_srccrop_y + max_crop_y / 2;
	}
	else // 1:1, 2:1, 4:1, custom, manual
	{
		int zoom_widget_pos_x = video_surface_infos->video_zoom_globalpos_x - widget_global_pos_x;

		if((zoom_widget_pos_x >= 0) && (zoom_widget_pos_x < video_surface_infos->sws_dst_width))
		{
			if(prev_crop_w > 0){
				int sws_prev_pos_x = video_surface_infos->sws_crop_x + prev_crop_w * zoom_widget_pos_x / video_surface_infos->sws_dst_width;
				mpxplay_debugf(DISPQT_DEBUGOUT_FFMPEG, "ZOOMX px:%d", sws_prev_pos_x);
				video_surface_infos->sws_crop_x = sws_prev_pos_x - video_surface_infos->sws_crop_w * zoom_widget_pos_x / video_surface_infos->sws_dst_width;
			}else{
				video_surface_infos->sws_crop_x = video_source_infos->sws_srccrop_x + max_crop_x * zoom_widget_pos_x / video_surface_infos->sws_dst_width;
			}
			if(video_surface_infos->sws_crop_x < video_source_infos->sws_srccrop_x)
				video_surface_infos->sws_crop_x = video_source_infos->sws_srccrop_x;
			else if(video_surface_infos->sws_crop_x > (video_source_infos->sws_srccrop_x + max_crop_x))
				video_surface_infos->sws_crop_x = video_source_infos->sws_srccrop_x + max_crop_x;
		}else{
			video_surface_infos->sws_crop_x = video_source_infos->sws_srccrop_x + max_crop_x / 2;
		}

		int zoom_widget_pos_y = video_surface_infos->video_zoom_globalpos_y - widget_global_pos_y;

		if((zoom_widget_pos_y >= 0) && (zoom_widget_pos_y < video_surface_infos->sws_dst_height))
		{
			if(prev_crop_h > 0){
				int sws_prev_pos_y = video_surface_infos->sws_crop_y + prev_crop_h * zoom_widget_pos_y / video_surface_infos->sws_dst_height;
				mpxplay_debugf(DISPQT_DEBUGOUT_FFMPEG, "ZOOMX px:%d", sws_prev_pos_y);
				video_surface_infos->sws_crop_y = sws_prev_pos_y - video_surface_infos->sws_crop_h * zoom_widget_pos_y / video_surface_infos->sws_dst_height;
			}else{
				video_surface_infos->sws_crop_y = video_source_infos->sws_srccrop_y + max_crop_y * zoom_widget_pos_y / video_surface_infos->sws_dst_height;
			}
			if(video_surface_infos->sws_crop_y < video_source_infos->sws_srccrop_y)
				video_surface_infos->sws_crop_y = video_source_infos->sws_srccrop_y;
			else if(video_surface_infos->sws_crop_y > (video_source_infos->sws_srccrop_y + max_crop_y))
				video_surface_infos->sws_crop_y = video_source_infos->sws_srccrop_y + max_crop_y;
		}else{
			video_surface_infos->sws_crop_y = video_source_infos->sws_srccrop_y + max_crop_y / 2;
		}

		if(zoom_value == zoom_val_ti)
			gcfg->video_zoom_type = DISPQT_VIDEOFULLSCREENZOOMTYPE_TOUCHINSIDE;
		else if(zoom_value == zoom_val_to)
			gcfg->video_zoom_type = DISPQT_VIDEOFULLSCREENZOOMTYPE_TOUCHOUTSIDE;
		else if(zoom_value == DISPQT_CONFIG_VIDEOZOOM_CENTER)
			gcfg->video_zoom_type = DISPQT_VIDEOFULLSCREENZOOMTYPE_1_1;

		mpxplay_debugf(DISPQT_DEBUGOUT_FFMPEG, "ZOOM cx:%d cy:%d", video_surface_infos->sws_crop_x, video_surface_infos->sws_crop_y);
	}

	mpxplay_debugf(DISPQT_DEBUGOUT_CROP, "sws_calculate war:%d var:%d n:%d d:%d sw:%d sh:%d ww:%d wh:%d z:%d zw:%d zh:%d cx:%d cy:%d cw:%d ch:%d",
			window_ar, video_ar, video_source_infos->video_aspect_ratio.num, video_source_infos->video_aspect_ratio.den, output_frame->width, output_frame->height,
			video_surface_infos->sws_dst_width, video_surface_infos->sws_dst_height, zoom_value, zoom_val_ti_w, zoom_val_ti_h, video_surface_infos->sws_crop_x, video_surface_infos->sws_crop_y,
			video_surface_infos->sws_crop_w, video_surface_infos->sws_crop_h);

	// --- target window calculate -----------------------------------------------------------------------------------------

	if(window_width < video_surface_infos->sws_dst_width)
	{
		video_surface_infos->sws_dst_pos_x = (video_surface_infos->sws_dst_width - window_width) / 2;
		video_surface_infos->sws_dst_width = window_width;
		video_surface_infos->videoout_surface_clearpic = true;
	}else{
		video_surface_infos->sws_dst_pos_x = 0;
	}
	if(window_height < video_surface_infos->sws_dst_height)
	{
		video_surface_infos->sws_dst_pos_y = (video_surface_infos->sws_dst_height - window_height) / 2;
		video_surface_infos->sws_dst_height = window_height;
		video_surface_infos->videoout_surface_clearpic = true;
	}else{
		video_surface_infos->sws_dst_pos_y = 0;
	}

	video_surface_infos->videoout_surface_recalc = false;

	return true;
}
//-------------------------------------------------------------------------------------------------------------------------------------

bool FFMpegVideoWidget::videowidget_ffmpeg_handle_refresh(unsigned int refresh_type, bool doRepaint)
{
	struct dispqt_video_source_info_s *video_source_infos = &this->video_source_infos;
	struct dispqt_video_surface_info_s *video_surface_infos = &this->video_surface_infos;

	if((refresh_type != DISPQT_VIDEOSURFACE_REFRESHTYPE_NONE) && (refresh_type != DISPQT_VIDEOSURFACE_REFRESHTYPE_CLOSE))
	{
		if(video_source_infos->update_crop)
		{
			this->video_sw_transform->videowidget_ffmpeg_filtered_crop_update(video_source_infos->check_autocrop);
			doRepaint = true;
		}
//		if(video_surface_infos->update_zoom)
//		{
//			doRepaint |= videowidget_ffmpeg_surface_zoom_calculate(video_surface_infos->updatezoom_center_x, video_surface_infos->updatezoom_center_y, video_surface_infos->updatezoom_direction);
//		}
		if(video_surface_infos->update_resize)
		{
			if(!video_source_infos->update_crop)
				this->video_sw_transform->videowidget_ffmpeg_filtered_crop_update(false);
			videowidget_ffmpeg_swsctx_calculate(video_surface_infos->updateresize_x, video_surface_infos->updateresize_y);
			doRepaint |= video_surface_infos->do_resize_repaint;
		}
//		if(video_surface_infos->update_move)
//		{
//			doRepaint |= videowidget_ffmpeg_surface_move_calculate(video_surface_infos->updatemove_x, video_surface_infos->updatemove_y);
//		}
		if(!doRepaint)
		{
			if(video_surface_infos->update_filter_trans)
				doRepaint = true;
			else if(video_surface_infos->update_filter_eq)
			{
				this->video_sw_transform->videowidget_ffmpeg_filtered_eq_update();
				this->videowidget_ffmpeg_filtered_surface_update(false);
			}
		}
	}

	if(refresh_type != DISPQT_VIDEOSURFACE_REFRESHTYPE_NONE)
	{
		video_source_infos->update_crop = false;
		video_source_infos->check_autocrop = false;
		//video_surface_infos->update_zoom = false;
		//video_surface_infos->updatezoom_direction = 0;
		video_surface_infos->update_resize = false;
		//video_surface_infos->update_move = false;
		//video_surface_infos->updatemove_x = 0;
		//video_surface_infos->updatemove_y = 0;
		video_surface_infos->update_filter_trans = false;
		video_surface_infos->update_filter_eq = false;
	}
	return doRepaint;
}

//------------------------------------------------------------------------------------------------------------------------

bool FFMpegVideoWidget::videowidget_ffmpeg_config_callback(unsigned int cfgcmd, mpxplay_infile_callback_t *cbk_func, void *passdata)
{
	bool is_video_open = false;
	mpxplay_debugf(DISPQT_DEBUGOUT_FFMPEG, "videowidget_ffmpeg_config_callback BEGIN %8.8X %8.8X %8.8X", cfgcmd, (int)cbk_func, (int)passdata);
	bool mutex_locked = this->mutex_video.tryLock(DISPQT_VIDEO_MUTEX_TIMEOUT);
	mpxplay_debugf(DISPQT_DEBUGOUT_FFMPEG, "videowidget_ffmpeg_config_callback LOCK %8.8X %8.8X %8.8X lock:%d", cfgcmd, (int)cbk_func, (int)passdata, (int)mutex_locked);
	if(this->video_decoder_worker)
		is_video_open = this->video_decoder_worker->videothread_ffmpeg_config_callback(cfgcmd, cbk_func, passdata);
	if(mutex_locked)
		this->mutex_video.unlock();
	mpxplay_debugf(DISPQT_DEBUGOUT_FFMPEG, "videowidget_ffmpeg_config_callback END %8.8X %8.8X %8.8X lock:%d", cfgcmd, (int)cbk_func, (int)passdata, (int)mutex_locked);
	return is_video_open;
}

void FFMpegVideoWidget::videowidget_ffmpeg_callback_call(int globalpos_x, int globalpos_y, unsigned long command, mpxp_ptrsize_t arg1, mpxp_ptrsize_t arg2)
{
	mpxplay_debugf(DISPQT_DEBUGOUT_FFMPEG, "videowidget_ffmpeg_callback_call BEGIN");
	bool mutex_locked = this->mutex_video.tryLock(DISPQT_VIDEO_MUTEX_TIMEOUT);
	if(this->video_decoder_worker)
		this->video_decoder_worker->videothread_ffmpeg_callback_call(globalpos_x, globalpos_y, command, arg1, arg2);
	if(mutex_locked)
		this->mutex_video.unlock();
	mpxplay_debugf(DISPQT_DEBUGOUT_FFMPEG, "videowidget_ffmpeg_callback_call END");
}

bool FFMpegVideoWidget::videowidget_ffmpeg_schedulerthread_suspend(void)
{
	return this->video_decoder_worker->decoderworker_video_scheduler_control(DISPQT_VIDEOSCHEDULER_CONTROL_REQ_SUSPEND);
}

void FFMpegVideoWidget::videowidget_ffmpeg_schedulerthread_resume(void)
{
	this->video_decoder_worker->decoderworker_video_scheduler_control(DISPQT_VIDEOSCHEDULER_CONTROL_RUN);
}

void FFMpegVideoWidget::videowidget_ffmpeg_frame_display(int video_index, unsigned int flags, AVFrame *videoout_frame, ffmpegvideo_subtitle_info_s *subtitle_infos, unsigned int refresh_type)
{
	struct dispqt_video_source_info_s *video_source_infos = &this->video_source_infos;
	struct dispqt_video_surface_info_s *video_surface_infos = &this->video_surface_infos;
	bool doRepaint = false;

	mpxplay_debugf(DISPQT_DEBUGOUT_RENDER, "videowidget_ffmpeg_frame_display BEGIN");

	if(refresh_type == DISPQT_VIDEOSURFACE_REFRESHTYPE_CLEAR) // means clear window
	{
		video_surface_infos->videoout_surface_recalc = true;
		video_surface_infos->videoout_surface_clearpic = true;
		doRepaint = true;
	}
	else if(refresh_type == DISPQT_VIDEOSURFACE_REFRESHTYPE_CLOSE) // means close rendering
	{
		this->video_sw_transform->videowidget_ffmpeg_videoctx_clear();
		this->video_sw_transform->videowidget_ffmpeg_swsctx_close();
		this->videowidget_ffmpeg_subtitle_close();
		video_surface_infos->videoout_surface_recalc = true;
		video_surface_infos->videoout_surface_clearpic = true;
		doRepaint = true;
		mpxplay_debugf(DISPQT_DEBUGOUT_FFMPEG, "videowidget_ffmpeg_frame_display CLOSE");
	}
	if(flags & DISPQT_FFMPGVIDEOCALLBACKFLAG_FIRSTFRAME) // means new file
	{
		video_surface_infos->video_outframe_last_width = 0;
		video_surface_infos->video_outframe_last_height = 0;
		this->video_sw_transform->videowidget_ffmpeg_swsctx_crop_auto_clear();
		mpxplay_debugf(DISPQT_DEBUGOUT_ACROP, "videowidget_ffmpeg_frame_display RESET2");
	}
	else if(!videoout_frame && (refresh_type == DISPQT_VIDEOSURFACE_REFRESHTYPE_RESET)) // means seeking
	{
		this->video_sw_transform->videowidget_ffmpeg_swsctx_crop_auto_reset();
		mpxplay_debugf(DISPQT_DEBUGOUT_ACROP, "videowidget_ffmpeg_frame_display RESET1");
	}

	if(videoout_frame) // there is a new decoded video frame
	{
		if(this->sub_title_infos)
		{
			mpxplay_video_ffmpeg_subtitleinfo_clear(this->sub_title_infos);
		}
		if(subtitle_infos)
		{
			if(!this->sub_title_infos)
				this->sub_title_infos = (ffmpegvideo_subtitle_info_s *)av_mallocz(sizeof(ffmpegvideo_subtitle_info_s));
			mpxplay_video_ffmpeg_subtitleinfo_copy(this->sub_title_infos, subtitle_infos);
		}
		if((video_source_infos->video_aspect_ratio.num != videoout_frame->sample_aspect_ratio.num) || (video_source_infos->video_aspect_ratio.den != videoout_frame->sample_aspect_ratio.den))
		{
			video_source_infos->video_aspect_ratio.num = videoout_frame->sample_aspect_ratio.num;
			video_source_infos->video_aspect_ratio.den = videoout_frame->sample_aspect_ratio.den;
			video_surface_infos->videoout_surface_recalc = true;
		}

		av_frame_free(&video_source_infos->ffmpeg_decoded_frame);
		video_source_infos->ffmpeg_decoded_frame = videoout_frame;
		videoout_frame = NULL;
		doRepaint = true;
	}

	doRepaint = this->videowidget_ffmpeg_handle_refresh(refresh_type, doRepaint);

	if(doRepaint)
	{
		if(this->main_window->gui_config->video_renderer_type == DISPQT_VIDEORENDERTYPE_WIDGET)
		{
			this->video_sw_transform->videowidget_ffmpeg_filtered_trans_update();
			this->videowidget_ffmpeg_filtered_surface_update(true);
		}
		else // TODO
		{
//			av_frame_free(&videoout_frame);
		}
	}
	mpxplay_debugf(DISPQT_DEBUGOUT_RENDER, "videowidget_ffmpeg_frame_display END");
}

#endif // MPXPLAY_LINK_ORIGINAL_FFMPEG
