//**************************************************************************
//*                     This file is part of the                           *
//*                MMC - Mpxplay Multimedia Commander                      *
//*                   The source code of MMC is                            *
//*        (C) copyright 1998-2019 by PDSoft (Attila Padar)                *
//*                http://mpxplay.sourceforge.net                          *
//*                  email: mpxplay@freemail.hu                            *
//**************************************************************************
//*  This program is distributed in the hope that it will be useful,       *
//*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *
//*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                  *
//*  Please contact with the author (with me) if you want to use           *
//*  or modify this source.                                                *
//**************************************************************************
//function: ffmpeg subtitle handling

//#define MPXPLAY_USE_DEBUGF 1
#define DISPQT_DEBUGOUT_SUBTITLE stdout

#include "moc_video_qt.h"

#define INFFMPG_MUTEX_ONTIMER_TIMEOUT     10 // used on 10ms timer

#define INFFMPG_SUBTITLE_DURATION_MUL  (1200)              // msec to AV_TIME_BASE, +20% duration for subtitles
#define INFFMPG_SUBTITLE_DISPLAY_GAP   (AV_TIME_BASE / 10) // min 0.1 sec gap between subtitles
#define INFFMPG_SUBTITLE_DISPLAY_MIN   (AV_TIME_BASE    )  // min 1 sec subtitle displaying (even if the next one is coming sooner)
#define INFFMPG_SUBTITLE_DISPLAY_HOLD  (AV_TIME_BASE * 2)  // hold for (min) 2 secs the subtitle displaying (even if the duration is less)
#define INFFMPG_SUBTITLE_DISPLAY_MAX   (AV_TIME_BASE * 10) // max 10 secs subtitle displaying

#ifdef MPXPLAY_LINK_ORIGINAL_FFMPEG

static unsigned int mpxplay_video_ffmpeg_subtitle_decode(ffmpegvideo_callback_list_s *cbelem, AVSubtitleRect *rect)
{
	ffmpegvideo_subtitle_info_s *sub_lines = cbelem->subtitle_infos;
	unsigned int lines = 0, textlen, flags, count, textcoding_type_source, textbuf_size;
	char *text, *next, *subtext_begin, *sb, strtmp[512];

	if(!sub_lines || !rect)
		return lines;

	pds_memset(sub_lines, 0, sizeof(*sub_lines));

	switch(rect->type){
		case SUBTITLE_TEXT:
			text = rect->text;
			break;
		case SUBTITLE_ASS:
			text = rect->ass;
			count = 9;
			do{
				next = pds_strchr(text, ',');
				if(!next)
					break;
				text = next + 1;
			}while(--count);
			break;
		default: return lines;
	}

	subtext_begin = text;

	//mpxplay_debugf(DISPQT_DEBUGOUT_SUBTITLE,"subtext_begin: %s", subtext_begin);

	text = subtext_begin;
	flags = 0;
	do{
		sub_lines->line_flags[lines] = flags;

		// search for end of line (multiply lines are merged in one subtitle, separated with newlines)
		next = NULL;
		textlen = 0;

		if(rect->type == SUBTITLE_ASS){
			next = text;
			while((next = pds_strchr(next, '\\'))){
				if((next[1] == 'N') || (next[1] == 'n')){
					textlen = next - text;
					next += 2;
					break;
				}
				next++;
			}
		}

		if(!next){
			if((next = pds_strstr(text, (char *)"\r\n")) || (next = pds_strstr(text, (char *)"\n\r"))){
				textlen = next - text;
				next += 2;
			}
			else if((next = pds_strchr(text, '\n'))){
				textlen = next - text;
				next++;
			}
		}

		if(!textlen){ // last line in this subtitle
			textlen = pds_strlen(text);
			if(!textlen)
				break;
		}
		if(textlen >= sizeof(strtmp))
			textlen = sizeof(strtmp) - 1;
		pds_strncpy(strtmp, text, textlen);
		strtmp[textlen] = 0;
		switch(rect->type){
			case SUBTITLE_TEXT: // TODO: not tested
				while((sb = pds_strchr(strtmp,'<'))){
					int ctrllen = 0;
					if(sb[2] == '>'){
						if(sb[1] == 'i'){  // "<i>"
							funcbit_enable(sub_lines->line_flags[lines], INFFMPEG_ASS_DECODED_FLAG_ITALIC); // TODO: flag changes inside line is not handled (only at begin and end)
							funcbit_enable(flags, INFFMPEG_ASS_DECODED_FLAG_ITALIC);
						}
						ctrllen = 3;
					}else if(sb[1] == '/' && sb[3] == '>'){
						if(sb[2] == 'i')  // "</i>"
							funcbit_disable(flags, INFFMPEG_ASS_DECODED_FLAG_ITALIC);
						ctrllen = 4;
					}

					if(ctrllen){
						pds_strcpy(sb, sb + ctrllen);
						textlen = pds_strlen(strtmp);
					}
				}
				break;
			case SUBTITLE_ASS:
				while((sb = pds_strchr(strtmp,'{'))){
					char *se;
					if(sb[1] == '\\' && sb[2] =='i')
						if(sb[3] == '1'){
							funcbit_enable(sub_lines->line_flags[lines], INFFMPEG_ASS_DECODED_FLAG_ITALIC);  // TODO: flag changes inside line is not handled (only at begin and end)
							funcbit_enable(flags, INFFMPEG_ASS_DECODED_FLAG_ITALIC);
						}else if(sb[3] == '0')
							funcbit_disable(flags, INFFMPEG_ASS_DECODED_FLAG_ITALIC);
					se = pds_strchr(sb, '}');
					if(se){
						if(se[1])
							pds_strcpy(sb, &se[1]);
						else // end of line
							sb[0] = 0;
					}
					textlen = pds_strlen(strtmp);
				}
				break;
		}
		textcoding_type_source = (cbelem->subtitle_pktlist_elem->flags >> MPXPLAY_PACKETLISTFLAG_TEXTTYPE_BITPOS) & MPXPLAY_PACKETLISTFLAG_TEXTTYPE_MASK;
		textbuf_size = textlen * 3; // max required space for char to UTF8
		sub_lines->line_texts[lines] = (char *)av_malloc(textbuf_size + 4);
		mpxplay_playlist_textconv_by_texttypes(MPXPLAY_TEXTCONV_TYPES_PUT(textcoding_type_source ,MPXPLAY_TEXTCONV_TYPE_MPXNATIVE), strtmp, textlen, sub_lines->line_texts[lines], textbuf_size + 2);
		lines++;
		text = next;
	}while(text && (lines < INFFMPEG_ASS_DECODED_MAX_LINES));
	sub_lines->nb_lines = lines;

	return lines;
}

void mpxplay_video_ffmpeg_subtitleinfo_clear(ffmpegvideo_subtitle_info_s *sub_lines)
{
	int i;
	if(!sub_lines || !sub_lines->nb_lines)
		return;
	for(i = 0; i < sub_lines->nb_lines; i++)
		av_freep(&sub_lines->line_texts[i]);
	pds_memset(sub_lines->line_flags, 0, sizeof(sub_lines->line_flags));
	sub_lines->nb_lines = 0;
}

void mpxplay_video_ffmpeg_subtitleinfo_copy(ffmpegvideo_subtitle_info_s *dest_lines, ffmpegvideo_subtitle_info_s *sub_lines)
{
	int i;
	if(!dest_lines || !sub_lines)
		return;
	dest_lines->nb_lines = sub_lines->nb_lines;
	pds_memcpy(dest_lines->line_flags, sub_lines->line_flags, sizeof(dest_lines->line_flags));
	for(i = 0; i < sub_lines->nb_lines; i++){
		unsigned int len = pds_strlen(sub_lines->line_texts[i]);
		if(len){
			len++;
			dest_lines->line_texts[i] = (char *)av_mallocz(len);
			if(dest_lines->line_texts[i])
				pds_memcpy(dest_lines->line_texts[i], sub_lines->line_texts[i], len);
		}else
			dest_lines->line_texts[i] = NULL;
	}
}


void mpxplay_decoderworker_subtitle_load_next_frame(ffmpegvideo_callback_list_s *cbelem)
{
	int retval_subtitle;
	int streamtype_index = MPXPLAY_STREAMTYPEINDEX_SUBTITLE;
	retval_subtitle = cbelem->infile_callback(cbelem->infile_passdata, MPXPLAY_INFILE_CBKCTRL_GET_PACKETLIST_ELEM, (mpxp_ptrsize_t)&cbelem->subtitle_pktlist_elem, streamtype_index, INFFMPG_MUTEX_ONTIMER_TIMEOUT);
	if((retval_subtitle >= 0) && cbelem->subtitle_pktlist_elem && (cbelem->subtitle_pktlist_elem->timestamp_us > 0) && (cbelem->subtitle_end_timestamp >= cbelem->subtitle_pktlist_elem->timestamp_us))
	{
		cbelem->subtitle_end_timestamp = cbelem->subtitle_pktlist_elem->timestamp_us - INFFMPG_SUBTITLE_DISPLAY_GAP;
		if((cbelem->subtitle_end_timestamp - cbelem->subtitle_begin_timestamp) < INFFMPG_SUBTITLE_DISPLAY_MIN)
			cbelem->subtitle_end_timestamp = cbelem->subtitle_begin_timestamp + INFFMPG_SUBTITLE_DISPLAY_MIN;
	}
}

void mpxplay_decoderworker_subtitle_decode(ffmpegvideo_callback_list_s *cbelem, mpxp_int64_t timestamp_curr_us)
{
	if(cbelem->subtitle_end_timestamp)
	{
		if(timestamp_curr_us >= cbelem->subtitle_end_timestamp)
		{
			mpxplay_video_ffmpeg_subtitleinfo_clear(cbelem->subtitle_infos);
			cbelem->subtitle_end_timestamp = 0;
		}
	}
	else if(cbelem->subtitle_pktlist_elem)
	{
		AVCodecContext *subtitle_codec_ctx = (AVCodecContext *)cbelem->subtitle_pktlist_elem->codec_ctx;
		int got_frame = 0;
		AVSubtitle subtitle_frame;

		pds_memset(&subtitle_frame, 0, sizeof(subtitle_frame));

		if(!cbelem->subtitle_infos)
			cbelem->subtitle_infos = (ffmpegvideo_subtitle_info_s *)av_mallocz(sizeof(ffmpegvideo_subtitle_info_s));
		else
			mpxplay_video_ffmpeg_subtitleinfo_clear(cbelem->subtitle_infos);

		if(subtitle_codec_ctx)
		{
			int retval = avcodec_decode_subtitle2(subtitle_codec_ctx, &subtitle_frame, &got_frame, (AVPacket *)(cbelem->subtitle_pktlist_elem->frame_pkt));
			if((retval >= 0) && got_frame && subtitle_frame.num_rects && subtitle_frame.rects)
			{
				if(mpxplay_video_ffmpeg_subtitle_decode(cbelem, subtitle_frame.rects[0]))  // TODO: possible multiply rects?
				{
					AVPacket *subtitle_packet = (AVPacket *)(cbelem->subtitle_pktlist_elem->frame_pkt);
					mpxp_int64_t duration;
					mpxplay_debugf(DISPQT_DEBUGOUT_SUBTITLE, "subtitle pts:%llu s:%d e:%d d:%llu %s", subtitle_packet->pts,
						subtitle_frame.start_display_time, subtitle_frame.end_display_time, subtitle_packet->duration, cbelem->subtitle_infos->line_texts[0]);
					cbelem->subtitle_begin_timestamp = cbelem->subtitle_pktlist_elem->timestamp_us;
					duration = subtitle_packet->duration * INFFMPG_SUBTITLE_DURATION_MUL;
					if(duration < INFFMPG_SUBTITLE_DISPLAY_HOLD)
						duration = INFFMPG_SUBTITLE_DISPLAY_HOLD;
					else if(duration > INFFMPG_SUBTITLE_DISPLAY_MAX)
						duration = INFFMPG_SUBTITLE_DISPLAY_MAX;
					cbelem->subtitle_end_timestamp = cbelem->subtitle_begin_timestamp + duration;
				}
			}
		}
		mpxplay_ffmpgdec_queuelistelem_clear(cbelem->subtitle_pktlist_elem);
		cbelem->subtitle_pktlist_elem = NULL;
		avsubtitle_free(&subtitle_frame);
	}
}

ffmpegvideo_subtitle_info_s *mpxplay_decoderworker_subtitle_get(ffmpegvideo_callback_list_s *cbelem, mpxp_int64_t timestamp_curr_us)
{
	ffmpegvideo_subtitle_info_s *subtitle = NULL;
	if(cbelem->subtitle_end_timestamp && (cbelem->subtitle_begin_timestamp <= timestamp_curr_us) && ((cbelem->subtitle_begin_timestamp + DISPQT_FFMPGVIDEO_SUBTITLE_MAX_VIDEO_DIFF_US) >= timestamp_curr_us))
	{
		subtitle = cbelem->subtitle_infos;
	}
	else if(((cbelem->subtitle_begin_timestamp + DISPQT_FFMPGVIDEO_SUBTITLE_MAX_VIDEO_DIFF_US) < timestamp_curr_us))
	{
		cbelem->subtitle_begin_timestamp = cbelem->subtitle_end_timestamp = 0;
	}
	return subtitle;
}

void mpxplay_decoderworker_subtitlepkt_clear(ffmpegvideo_callback_list_s *cbelem)
{
	if(cbelem->subtitle_pktlist_elem)
	{
		mpxplay_ffmpgdec_queuelistelem_clear(cbelem->subtitle_pktlist_elem);
		cbelem->subtitle_pktlist_elem = NULL;
	}
}

#if 0
void mpxplay_video_ffmpeg_avsubtitle_copy(AVSubtitle *avsub_dest, AVSubtitle *avsub_src)
{
	if(!avsub_dest || !avsub_src)
		return;
	pds_memcpy(avsub_dest, avsub_src, sizeof(AVSubtitle));
	if(avsub_src->num_rects && avsub_src->rects)
	{
		int i;
		avsub_dest->rects = (AVSubtitleRect **)av_mallocz(avsub_src->num_rects * sizeof(AVSubtitleRect *));
		for(i = 0; i < avsub_dest->num_rects; i++)
		{
			unsigned int len;
			avsub_dest->rects[i] = (AVSubtitleRect *)av_mallocz(sizeof(AVSubtitleRect));
			if(!avsub_dest->rects[i])
				break;
			//pds_memcpy(avsub_dest->rects[i], avsub_src->rects[i], sizeof(AVSubtitleRect)); //TODO
			avsub_dest->rects[i]->type = avsub_src->rects[i]->type;
			len = pds_strlen(avsub_src->rects[i]->text);
			if(len){
				avsub_dest->rects[i]->text = (char *)av_malloc(len + 1);
				pds_memcpy(avsub_dest->rects[i]->text, avsub_src->rects[i]->text, len + 1);
			}
			len = pds_strlen(avsub_src->rects[i]->ass);
			if(len){
				avsub_dest->rects[i]->ass = (char *)av_malloc(len + 1);
				pds_memcpy(avsub_dest->rects[i]->ass, avsub_src->rects[i]->ass, len + 1);
			}
		}
	}
}
#endif

#endif // MPXPLAY_LINK_ORIGINAL_FFMPEG
